/*
  
  This file is part of ERIKYYYPHONE
  Copyright (C) 1996 Erik Thiele
  see README for details and contact adresses.

  This program is free software; you can redistribute it and/or modify
  it under the terms of the GNU General Public License as published by
  the Free Software Foundation; either version 2 of the License, or
  (at your option) any later version.
  
  This program is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.
  
  You should have received a copy of the GNU General Public License
  along with this program; if not, write to the Free Software
  Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.

  */

#include "GUI.H"

#include "warter.H"
#include "fatal.H"
#include <ncurses.h>
#include <stdio.h>
#include <string.h>
#include <stdlib.h>
#include <signal.h>
#include <errno.h>
#include <unistd.h>

static int inited=0;
static int cx;
static int cy;
static int minx;
static int miny;
static int cursoranaus;
static int reversemode;

int GUI_KEYDELETE = KEY_MAX + 6;
int GUI_KEYUP     = KEY_MAX + 7;
int GUI_KEYDOWN   = KEY_MAX + 8;
int GUI_KEYLEFT   = KEY_MAX + 9;
int GUI_KEYRIGHT  = KEY_MAX + 10;

void fatalfehler(const char *const s)
{
  if (inited) GUI_deinit();
  fprintf(stderr,"\n\n");
  fprintf(stderr,"*** Fatal error, program terminated, core dumped ***\n\n");
  fprintf(stderr,"%s\n\n",s);
  raise(SIGSEGV);
};

void fatalfehlererrno(const char *const s)
{
  if (inited) GUI_deinit();
  fprintf(stderr,"\n\n");
  fprintf(stderr,"*** Fatal error, program terminated, core dumped ***\n\n");
  fprintf(stderr,"%s\n",s);
  fprintf(stderr,"errno: %s\n\n",strerror(errno));
  raise(SIGSEGV);
};

void GUI_init(const int minimum_x,const int minimum_y)
{
  if (inited) fatalfehler ("GUI_init: GUI is already initialized");
  WARTER_add(WARTER_SIGNAL,SIGWINCH); // Wenn du die Reihenfolge dieser Befehle 
                                      // aenderst, wirst du gekillt !
  minx=minimum_x;
  miny=minimum_y;
  
  if (initscr()!=stdscr) fatalfehler ("GUI_init: initscr() failed");
  if (noecho()==ERR) fatalfehler ("GUI_init: noecho() failed");
  timeout(0);
  if (nonl()==ERR) fatalfehler ("GUI_init: nonl() failed");
  if (cbreak()==ERR) fatalfehler ("GUI_init: cbreak() failed");
  if (keypad(stdscr,TRUE)==ERR) fatalfehler ("GUI_init: keypad(stdscr,TRUE) failed");
  if (notimeout(stdscr,FALSE)==ERR) fatalfehler ("GUI_init: notimeout(stdscr,FALSE) failed");
  if (leaveok(stdscr,TRUE)==ERR) fatalfehler ("GUI_init: leaveok(stdscr,TRUE) failed");

  GUI_reverse(0);
  GUI_cursor_setpos(0,0);
  GUI_cursor_visibility(1);
  GUI_putbox(0,0,GUI_getanzx()-1,GUI_getanzy()-1,' ');
  WARTER_add(WARTER_READ,0);
  inited=1;
};

void GUI_deinit()
{
  if (!inited) fatalfehler("GUI_deinit: i won't deinitialize a not initialized GUI");
  inited=0;

  /* HISTORY REPORT:
     On Sat Sep 28 17:25:05 MET DST 1996
     Erik Thiele alias ERIKYYY alias me
     spent 6 hours only to find out that
     inited=0 has to come before
     WARTER_signal_del(SIGWINCH);
     if it doesn't, an endless loop results. */

  if (endwin()==ERR) fatalfehler ("GUI_deinit: endwin() failed");
  WARTER_del(WARTER_SIGNAL,SIGWINCH);
  WARTER_del(WARTER_READ,0);
  printf("\n\n\n\n");
  fflush(stdout);
};

void GUI_update()
{
  if (WARTER_signal_aktiv_und_reset(SIGWINCH)) {
    endwin();
    refresh();

    GUI_cursor_setpos(0,0);
    GUI_putbox(0,0,GUI_getanzx()-1,GUI_getanzy()-1,' ');
    GUI_reverse(reversemode);
    WARTER_reenter=1;
  };
  
  int x,y;
  x = (cx<COLS) ? cx : (COLS-1);
  y = (cy<LINES) ? cy : (LINES-1);
  if (move(y,x)==ERR) fatalfehler("GUI_update: move failed");
  curs_set ( ((cx<COLS)&&(cy<LINES)) ? (cursoranaus ? TRUE : FALSE) : FALSE );
  // kein fatalfehlercheck, da wenn nicht geklappt hat, dann auch egal...

  if (refresh()==ERR) fatalfehler("GUI_update: refresh() failed");
};

void GUI_putchar(const int x,const int y,const int c)
{
  chtype i;
  switch(c)
    {
    case GUI_HLINE: i=ACS_HLINE;break;
    case GUI_VLINE: i=ACS_VLINE;break;
    case GUI_ULCORNER: i=ACS_ULCORNER;break;
    case GUI_URCORNER: i=ACS_URCORNER;break;
    case GUI_LLCORNER: i=ACS_LLCORNER;break;
    case GUI_LRCORNER: i=ACS_LRCORNER;break;
    default:
      i=c;
      break;
    };

  if ((x<0)||(y<0)||(x>=GUI_getanzx())||(y>=GUI_getanzy()))
    fatalfehler("GUI_putchar(...): position out of range");
  if ((x<COLS)&&(y<LINES))
    mvaddch(y,x,i);
};

void GUI_cursor_visibility(const int anaus)
{
  cursoranaus=anaus;
};

void GUI_cursor_setpos(const int x,const int y)
{
  if ((x<0)||(y<0)||(x>=GUI_getanzx())||(y>=GUI_getanzy()))
    fatalfehler("GUI_cursor_setpos(...): position out of range");
  cx=x;
  cy=y;
};

int GUI_getanzx()
{
  return( (minx>COLS) ? minx : COLS);
};

int GUI_getanzy()
{
  return( (miny>LINES) ? miny : LINES);
};

void GUI_putstring(const int x,const int y,const char *const s)
{
  for (unsigned int a=0;a<strlen(s);a++) GUI_putchar(x+a,y,s[a]);
};

void GUI_putrahmen(const int x1,const int y1,const int x2,const int y2)
{
  GUI_puthline(x1+1,y1,x2-1);
  GUI_puthline(x1+1,y2,x2-1);
  GUI_putvline(x1,y1+1,y2-1);
  GUI_putvline(x2,y1+1,y2-1);

  GUI_putchar(x1,y1,GUI_ULCORNER);
  GUI_putchar(x1,y2,GUI_LLCORNER);
  GUI_putchar(x2,y1,GUI_URCORNER);
  GUI_putchar(x2,y2,GUI_LRCORNER);
};

void GUI_putbox(const int x1,const int y1,const int x2,const int y2,const int c)
{
  for (int a=x1;a<=x2;a++)
    for (int b=y1;b<=y2;b++)
      GUI_putchar(a,b,c);
};

void GUI_putcenterstring(const int x,const int y,const char *const s)
{
  GUI_putstring(x-int(strlen(s)>>1),y,s);
};

void GUI_puthline(const int x1,const int y1,const int x2)
{
  for (int a=x1;a<=x2;a++) {
    GUI_putchar(a,y1,GUI_HLINE);
  };
};

void GUI_putvline(const int x1,const int y1,const int y2)
{  
  for (int a=y1;a<=y2;a++) {
    GUI_putchar(x1,a,GUI_VLINE);
  };
};

int GUI_getch(int *const zeichen)
{
  int e=getch();
  switch(e)
    {
    case ERR: return(0);
    case KEY_BACKSPACE:
    case KEY_DC: e=GUI_KEYDELETE;break;
    case KEY_UP: e=GUI_KEYUP;break;
    case KEY_DOWN: e=GUI_KEYDOWN;break;
    case KEY_LEFT: e=GUI_KEYLEFT;break;
    case KEY_RIGHT: e=GUI_KEYRIGHT;break;
    };
  *zeichen=e;
  return(1);
};

void GUI_reverse(const int anaus)
{
  reversemode=anaus;
  if (reversemode) attron(A_REVERSE); else attroff(A_REVERSE);
};
