/*
  
  This file is part of ERIKYYYPHONE
  Copyright (C) 1996 Erik Thiele
  see README for details and contact adresses.

  This program is free software; you can redistribute it and/or modify
  it under the terms of the GNU General Public License as published by
  the Free Software Foundation; either version 2 of the License, or
  (at your option) any later version.
  
  This program is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.
  
  You should have received a copy of the GNU General Public License
  along with this program; if not, write to the Free Software
  Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.

  */

#ifndef __MSG_H
#define __MSG_H

#include "allgtypedef.H"
#include "buffer.H"
#include "warter.H"


/*******************************************************************
  
                         common utilities

*******************************************************************/

#ifdef MSG_NEED_SPECIAL_UTILITIES
void set_socket_reuse(int handle);
void set_socket_linger(int handle);
void set_socket_nonblock(int handle);
void create_socket(int *handle);
//void create_UDP_socket(int *handle);
void bind_socket(int handle,uint16 portnum);
void listen_socket(int handle);
void accept_socket(int handle,int *neuhandle,int32 *herkunftip);
int connect_socket(int handle,int32 zielip,uint16 portnum);
#endif

void MSG_ip2string(char **str,int32 ip); // str muss danach wieder freigegeben werden !!!
int MSG_string2ip(char *str,int32 *ip);


/*******************************************************************
  
                             MSG_server

*******************************************************************/

class MSG_server {
public:
  MSG_server(uint16 portnum); // portnum in NETWORK BYTE ORDER !!
  ~MSG_server();

  int acceptnew(int *neuhandle,int32 *herkunftip);  // herkunftip in NETWORK BYTE ORDER !!

private:
  int handle;
};


/*******************************************************************
  
                             MSG_client

*******************************************************************/

class MSG_client {
public:
  enum statustype { CONNECTING, ERROR, CONNECTED };

  MSG_client(int32 zielip_,uint16 portnum_); // portnum_, zielip_ in NETWORK BYTE ORDER !!
  ~MSG_client();

  statustype connectready(int *neuhandle);

private:
  int32 zielip;
  uint16 portnum;
  int handle;
  statustype status;
};


/*******************************************************************
  
                          MSG_communicator

*******************************************************************/

class MSG_communicator {
public:
  MSG_communicator(int handle);
  ~MSG_communicator();

  int msgavail();
  void recvmsg(uint8 *msgart,char **buf,uint16 *len); // achtung ! free() nicht vergessen.
  void sendmsg(uint8 msgart,char *buf,uint16 len);
  // spaeter buf als buffer klasse, damit nicht immer dieser free() kack.


  int handleevent();  // !!! stellt evtl. mehrere msgs zur verfuegung !!!
  // wenn aber via recvmsg eine geholt wird und es ist danach noch eine da,
  // wird ein WARTER_reenter=1; ausgefuehrt.
  // daher kein while(msgavail()) recvmsg() noetig.

private:
  void write_wait(int anaus);

  int handle;
  int write_wait_stat;

  buffer *outbuffer,*inbuffer;
};


/*******************************************************************
  
                           MSG_UDP_server

*******************************************************************/

/*class MSG_UDP_server {
public:
  MSG_UDP_server(uint16 portnum); // portnum in NETWORK BYTE ORDER !!
  ~MSG_UDP_server();

  int getmsg(char *buf,int *size,int32 *herkunft);
  // buf wird gefuellt. size wird geaendert auf tatsaechliche groesse, herkunft wird gesetzt
  // 1=empfangen 0=nichts empfangen

private:
  int handle;
};*/

#endif
