/*
  
  This file is part of ERIKYYYPHONE
  Copyright (C) 1996 Erik Thiele
  see README for details and contact adresses.

  This program is free software; you can redistribute it and/or modify
  it under the terms of the GNU General Public License as published by
  the Free Software Foundation; either version 2 of the License, or
  (at your option) any later version.
  
  This program is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.
  
  You should have received a copy of the GNU General Public License
  along with this program; if not, write to the Free Software
  Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.

  */

#include "warter.H"

#include "erikyyylist.H"
#include "fatal.H"
#include <unistd.h>
#include <signal.h>
#include <sys/time.h>
#include <string.h>
#include <stdlib.h>
#include <fcntl.h>
#include <errno.h>

#define ANZSIG 100
static liste<int> *readlist,*writelist,*signallist;
static int breakpipe[2];
static int breakpipebenutzt;
static char signal_pending[ANZSIG];
static int num;
int WARTER_reenter;


void WARTER_init()
{
  readlist=new liste<int>();
  writelist=new liste<int>();
  signallist=new liste<int>();
  if (pipe(breakpipe)<0)fatalfehlererrno("WARTER_init(): pipe(...) failed");
  if (fcntl(breakpipe[0],F_SETFL,O_NONBLOCK)<0)fatalfehlererrno("WARTER_init(): fcntl nonblock failed (0)");
  breakpipebenutzt=0;
  for (int i=0;i<ANZSIG;i++)signal_pending[i]=0;
  if ((num=getdtablesize())<0) fatalfehlererrno("WARTER_warte(): getdtablesize()<0");
  WARTER_reenter=0;
};


void WARTER_deinit()
{
  if (readlist->first) fatalfehler("WARTER_deinit: readlist not empty.");
  delete readlist;
  if (writelist->first) fatalfehler("WARTER_deinit: writelist not empty.");
  delete writelist;
  if (signallist->first) fatalfehler("WARTER_deinit: signallist not empty.");
  delete signallist;
  if (close(breakpipe[1])<0) fatalfehlererrno("WARTER_deinit: close failed (1)");
  if (close(breakpipe[0])<0) fatalfehlererrno("WARTER_deinit: close failed (0)");
};


static void signalhandler(int sig)
{
  signal_pending[sig]=1;

  if (!breakpipebenutzt) {
    char a;
    if (write(breakpipe[1],&a,1)<0)fatalfehlererrno("signalhandler: write failed");
    breakpipebenutzt=1;
  };
};


void WARTER_warte()
{
  fd_set readfds,writefds;
  sigset_t set,signal_mask;
  char signal_checking[ANZSIG];
  struct sigaction act;
  int i,erg,erg2;
  liste_elem<int> *where;
  timeval *timeout,nulltime={0,0};

  // ***** TIMEOUT JE NACH REENTER SETZEN
  if (WARTER_reenter) {
    timeout=&nulltime;
    WARTER_reenter=0;
  }else{
    timeout=NULL;
  };

  // ***** READ HANDLES AKTIVIEREN
  FD_ZERO(&readfds);
  for (where=readlist->first;where;where=where->next) {
    FD_SET(*(where->objptr),&readfds);
  };
  
  // ***** BREAKPIPE AKTIVIEREN    
  breakpipebenutzt=0;
  FD_SET(breakpipe[0],&readfds);
  
  // ***** WRITE HANDLES AKTIVIEREN
  FD_ZERO(&writefds);
  for (where=writelist->first;where;where=where->next) {
    FD_SET(*(where->objptr),&writefds);
  };
  
  // ***** SIGNALE AKTIVIEREN
  // -- signal_mask leeren
  if (sigemptyset(&(signal_mask))<0)
    fatalfehlererrno("WARTER_warte(): sigemptyset(...)<0  (8)");
  // -- signal_checking leeren
  for (i=0;i<ANZSIG;i++) signal_checking[i]=0;
  // -- signal_checking fuellen anhand der signallist
  for (where=signallist->first;where;where=where->next) {
    signal_checking[*(where->objptr)]=1;
  };
  // -- signal_mask fuellen anhand von signal_checking
  for (i=0;i<ANZSIG;i++) if (signal_checking[i]) {
    if (sigaddset(&signal_mask,i)<0) fatalfehlererrno("WARTER_warte(): sigaddset(...)<0");
  };
  // -- signal_handler aktivieren anhand von signal_checking und signal_mask
  for (i=0;i<ANZSIG;i++) if (signal_checking[i]) {
    act.sa_handler=signalhandler;
    act.sa_mask=signal_mask;
    act.sa_flags=0;
    act.sa_restorer=NULL;
    while (sigaction(i,&act,NULL)<0) if (errno!=EINTR)
      fatalfehlererrno("WARTER_warte(): sigaction(...)<0");
    if (sigemptyset(&set)<0) fatalfehlererrno("WARTER_warte(): sigemptyset(...)<0 (2)");
    if (sigaddset(&set,i)<0) fatalfehlererrno("WARTER_warte(): sigaddset(...)<0 (1)");
    while (sigprocmask(SIG_UNBLOCK,&set,NULL)<0) if (errno!=EINTR)
      fatalfehlererrno("WARTER_warte(): sigprocmask(...)<0 (1)");
  };
  
  // ***** HAUPTSCHLEIFE.
  do {
    erg=select(num,&readfds,&writefds,NULL,timeout);
    if ((erg<0)&&(errno!=EINTR)) fatalfehlererrno("WARTER_warte(): select(...)<0");      
  }while(erg<0);
  
  // ***** SIGNALE DEAKTIVIEREN
  for (i=0;i<ANZSIG;i++) if (signal_checking[i]) {
    if (sigemptyset(&set)<0) fatalfehlererrno("WARTER_warte(): sigemptyset(...)<0 (3)");
    if (sigaddset(&set,i)<0) fatalfehlererrno("WARTER_warte(): sigaddset(...)<0 (2)");
    while (sigprocmask(SIG_BLOCK,&set,NULL)<0) if (errno!=EINTR)
      fatalfehlererrno("WARTER_warte(): sigprocmask(...)<0 (2)");
  };
  
  // ***** BREAKPIPE CLEAREN
  do{
    if ((erg2=read(breakpipe[0],&i,1))<0)
      if (errno!=EAGAIN) fatalfehlererrno("WARTER_warte: breakpipe read failed");
  }while(erg2>0);
};


void WARTER_add(WarterType art,int i)
{
  liste_elem<int> *a;
  liste<int> *l;

#ifdef kickwarn
  l=NULL;
#endif

  switch(art) 
    {
    case WARTER_SIGNAL:
      {
	sigset_t set;
	if ((i<0)||(i>=ANZSIG)) fatalfehler("WARTER_add: SIGNAL out of range");
	if (sigemptyset(&set)<0) fatalfehlererrno("WARTER_add(): sigemptyset(...)<0 (3)");
	if (sigaddset(&set,i)<0) fatalfehlererrno("WARTER_add(): sigaddset(...)<0 (2)");
	if (sigprocmask(SIG_BLOCK,&set,NULL)<0)
	  fatalfehlererrno("WARTER_add(): sigprocmask(...)<0 (2)");
	
	l=signallist;
	break;
      };
    case WARTER_READ:
      {
	l=readlist;
	break;
      };
    case WARTER_WRITE:
      {
	l=writelist;
	break;
      };
    default:
      {
	fatalfehler("WARTER_add(...): art != READ != WRITE != SIGNAL");
      };
    };
  for(a=l->first;a;a=a->next)
    if (*(a->objptr)==i)fatalfehler("WARTER_add(...): schon in liste");
  l->addfront(new int(i));
};


void WARTER_del(WarterType art,int i)
{
  liste_elem<int> *a;
  liste<int> *l;

#ifdef kickwarn
  l=NULL;
#endif

  switch(art) 
    {
    case WARTER_SIGNAL:
      {
	if ((i<0)||(i>=ANZSIG)) fatalfehler("WARTER_del: SIGNAL out of range");
	l=signallist;
	break;
      };
    case WARTER_READ:
      {
	l=readlist;
	break;
      };
    case WARTER_WRITE:
      {
	l=writelist;
	break;
      };
    default:
      {
	fatalfehler("WARTER_del(...): art != READ != WRITE != SIGNAL");
      };
    };

  a=l->first;
  int deleted=0;
  while(a) {
    if (*(a->objptr)==i) {
      l->del_and_next(&a);
      deleted=1;
    }else{
      a=a->next;
    };
  };
  if (!deleted) fatalfehler("WARTER_del(...): nicht zum loeschen vorhanden");
};


int WARTER_signal_aktiv_und_reset(int signum)
{
  if ((signum<0)||(signum>=ANZSIG)) fatalfehler("WARTER_signal_aktiv: unknown signal");
  if (signal_pending[signum]) {
    signal_pending[signum]=0;
    return(1);
  }else{
    return(0);
  };
};


int WARTER_handle_aktiv(WarterType mode,int handle)
{
  timeval nulltime={0,0};
  fd_set fds;

  FD_ZERO (&fds);
  FD_SET (handle,&fds);

  if (mode==WARTER_READ) {
    if (select(num,&fds,NULL,NULL,&nulltime)<0)
      fatalfehlererrno("warter.H: int WARTER_handle_aktiv(WARTER_READ,int handle): select(...)<0");
  }else if (mode==WARTER_WRITE) {
    if (select(num,NULL,&fds,NULL,&nulltime)<0)
      fatalfehlererrno("warter.H: int WARTER_handle_aktiv(WARTER_WRITE,int handle): select(...)<0");
  }else{
    fatalfehler("WARTER_handle_aktiv: mode !=READ != WRITE");
  };
  return(FD_ISSET(handle,&fds));
};
